@@ -40,6 +40,12 @@ typedef enum {
     json_create_invalid_user_json,
     /* User gave us an undefined value from a user subroutine. */
     json_create_undefined_return_value,
+    /* Rejected non-ASCII, non-character string in strict mode. */
+    json_create_non_ascii_byte,
+    /* Rejected scalar reference in strict mode. */
+    json_create_scalar_reference,
+    /* Rejected non-finite number in strict mode. */
+    json_create_non_finite_number,
 }
 json_create_status_t;
 
@@ -90,6 +96,8 @@ typedef struct json_create {
     unsigned int downgrade_utf8 : 1;
     /* Output may contain invalid UTF-8. */
     unsigned int utf8_dangerous : 1;
+    /* Strict mode, reject lots of things. */
+    unsigned int strict : 1;
 }
 json_create_t;
 
@@ -133,6 +141,9 @@ static int (* json_create_error_handler) (const char * file, int line_number, co
 	case json_create_unicode_bad_utf8:			\
 	case json_create_invalid_user_json:			\
 	case json_create_undefined_return_value:		\
+	case json_create_non_ascii_byte:			\
+	case json_create_scalar_reference:			\
+	case json_create_non_finite_number:			\
 	    break;						\
 	    							\
 	    /* All other exceptions are our bugs. */		\
@@ -217,10 +228,10 @@ add_char (json_create_t * jc, unsigned char c)
 
 /* Add a nul-terminated string to "jc", up to the nul byte. This
    should not be used unless it's strictly necessary, prefer to use
-   "add_str_len" instead. This is not intended to be Unicode-safe, it
-   is only to be used for strings which we know do not need to be
-   checked for Unicode validity (for example sprintf'd numbers or
-   something). Basically, don't use this. */
+   "add_str_len" instead. Basically, don't use this. This is not
+   intended to be Unicode-safe, it is only to be used for strings
+   which we know do not need to be checked for Unicode validity (for
+   example sprintf'd numbers or something). */
 
 static INLINE json_create_status_t
 add_str (json_create_t * jc, const char * s)
@@ -386,6 +397,91 @@ static jump_t jump[0x100] = {
     UT4,UT4,UT4,UT4,UT4,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,BAD,
 };
 
+/* Need this twice, once within the ASCII handler and once within the
+   Unicode handler. */
+
+#define ASCII \
+	case CTL:\
+	    CALL (add_one_u (jc, (unsigned int) c));\
+	    i++;\
+	    break;\
+\
+	case BSX:\
+	    ADD ("\\b");\
+	    i++;\
+	    break;\
+\
+	case HTX:\
+	    ADD ("\\t");\
+	    i++;\
+	    break;\
+\
+	case NLX:\
+	    ADD ("\\n");\
+	    i++;\
+	    break;\
+\
+	case NPX:\
+	    ADD ("\\f");\
+	    i++;\
+	    break;\
+\
+	case CRX:\
+	    ADD ("\\r");\
+	    i++;\
+	    break;\
+\
+	case ASC:\
+	    CALL (add_char (jc, c));\
+	    i++;\
+	    break;\
+\
+	case QUO:\
+	    ADD ("\\\"");\
+	    i++;\
+	    break;\
+\
+	case FSL:\
+	    if (jc->escape_slash) {\
+		ADD ("\\/");\
+	    }\
+	    else {\
+		CALL (add_char (jc, c));\
+	    }\
+	    i++;\
+	    break;\
+\
+	case BSL:\
+	    ADD ("\\\\");\
+	    i++;\
+	    break;
+
+
+static INLINE json_create_status_t
+json_create_add_ascii_key_len (json_create_t * jc, const unsigned char * key, STRLEN keylen)
+{
+    int i;
+
+    CALL (add_char (jc, '"'));
+    for (i = 0; i < keylen; ) {
+	char c;
+	c = key[i];
+	switch (jump[c]) {
+
+	ASCII;
+
+	default:
+	    json_create_user_message (jc, json_create_non_ascii_byte,
+				      "Non-ASCII byte in non-utf8 string: %X",
+				      key[i]);
+	    return json_create_non_ascii_byte;
+	}
+    }
+    CALL (add_char (jc, '"'));
+    return json_create_ok;
+}
+
+
 /* Add a string to the buffer with quotes around it and escapes for
    the escapables. */
 
@@ -401,60 +497,7 @@ json_create_add_key_len (json_create_t * jc, const unsigned char * key, STRLEN k
 
 	switch (jump[c]) {
 
-	case CTL:
-	    CALL (add_one_u (jc, (unsigned int) c));
-	    i++;
-	    break;
-
-	case BSX:
-	    ADD ("\\b");
-	    i++;
-	    break;
-
-	case HTX:
-	    ADD ("\\t");
-	    i++;
-	    break;
-
-	case NLX:
-	    ADD ("\\n");
-	    i++;
-	    break;
-
-	case NPX:
-	    ADD ("\\f");
-	    i++;
-	    break;
-
-	case CRX:
-	    ADD ("\\r");
-	    i++;
-	    break;
-
-	case ASC:
-	    CALL (add_char (jc, c));
-	    i++;
-	    break;
-
-	case QUO:
-	    ADD ("\\\"");
-	    i++;
-	    break;
-
-	case FSL:
-	    if (jc->escape_slash) {
-		ADD ("\\/");
-	    }
-	    else {
-		CALL (add_char (jc, c));
-	    }
-	    i++;
-	    break;
-
-	case BSL:
-	    ADD ("\\\\");
-	    i++;
-	    break;
+	ASCII;
 
 	case BAD:
 	    BADUTF8;
@@ -562,16 +605,19 @@ json_create_add_string (json_create_t * jc, SV * input)
 {
     char * istring;
     STRLEN ilength;
-    /*
-      "jc->unicode" is true if Perl says that anything in the whole of
-      the input to "json_create" is a "SvUTF8" scalar.
-    */
     istring = SvPV (input, ilength);
     if (SvUTF8 (input)) {
-	/* We have to force everything in the whole output to
+	/* "jc->unicode" is true if Perl says that anything in the
+	   whole of the input to "json_create" is a "SvUTF8"
+	   scalar. We have to force everything in the whole output to
 	   Unicode. */
 	jc->unicode = 1;
     }
+    else if (jc->strict) {
+	/* Backtrace fall through, remember to check the caller's line. */
+	return json_create_add_ascii_key_len (jc, (unsigned char *) istring,
+					      (STRLEN) ilength);
+    }
     /* Backtrace fall through, remember to check the caller's line. */
     return json_create_add_key_len (jc, (unsigned char *) istring,
 				    (STRLEN) ilength);
@@ -787,13 +833,14 @@ json_create_call_to_json (json_create_t * jc, SV * cv, SV * r)
     
     PUSHMARK (SP);
     //https://metacpan.org/source/AMBS/Math-GSL-0.35/swig/gsl_typemaps.i#L482
-    XPUSHs (sv_2mortal(newRV(r)));
+    XPUSHs (sv_2mortal (newRV (r)));
     PUTBACK;
     call_sv (cv, 0);
     json = POPs;
     SvREFCNT_inc (json);
     FREETMPS;
     LEAVE;  
+
     if (! SvOK (json)) {
 	/* User returned an undefined value. */
 	SvREFCNT_dec (json);
@@ -844,6 +891,11 @@ json_create_add_float (json_create_t * jc, SV * sv)
 	    CALL (json_create_call_to_json (jc, jc->non_finite_handler, sv));
 	}
 	else {
+	    if (jc->strict) {
+		json_create_user_message (jc, json_create_non_finite_number,
+					  "Non-finite number in input");
+		return json_create_non_finite_number;
+	    }
 	    if (isnan (fv)) {
 		ADD ("\"nan\"");
 	    }
@@ -910,17 +962,25 @@ json_create_add_object (json_create_t * jc, HV * input_hv)
 	/* The following is necessary because "hv_iternextsv" doesn't
 	   tell us whether the key is "SvUTF8" or not. */
 	he = hv_iternext (input_hv);
-	if (HeUTF8 (he)) {
-	    jc->unicode = 1;
-	}
 	key = hv_iterkey (he, & keylen);
 	value = hv_iterval (input_hv, he);
 
 	/* Write the information into the buffer. */
 
 	COMMA;
-	CALL (json_create_add_key_len (jc, (const unsigned char *) key,
-				       (STRLEN) keylen));
+	if (HeUTF8 (he)) {
+	    jc->unicode = 1;
+	    CALL (json_create_add_key_len (jc, (const unsigned char *) key,
+					   (STRLEN) keylen));
+	}
+	else if (jc->strict) {
+	    CALL (json_create_add_ascii_key_len (jc, (unsigned char *) key,
+						 (STRLEN) keylen));
+	}
+	else {
+	    CALL (json_create_add_key_len (jc, (const unsigned char *) key,
+					   (STRLEN) keylen));
+	}
 	CALL (add_char (jc, ':'));
 	CALL (json_create_recursively (jc, value));
     }
@@ -967,6 +1027,14 @@ json_create_handle_unknown_type (json_create_t * jc, SV * r)
     return json_create_ok;
 }
 
+#define STRICT_NO_SCALAR						\
+    if (jc->strict) {							\
+	json_create_user_message (jc, json_create_scalar_reference,	\
+				  "Scalar reference rejected");		\
+	return json_create_scalar_reference;				\
+    }
+
+
 static INLINE json_create_status_t
 json_create_handle_ref (json_create_t * jc, SV * input)
 {
@@ -974,6 +1042,10 @@ json_create_handle_ref (json_create_t * jc, SV * input)
     SV * r;
     r = SvRV (input);
     t = SvTYPE (r);
+#if 0
+    fprintf (stderr, "%s:%d: type is %d\n", __FILE__, 
+	     __LINE__, t);
+#endif /* 0 */
     switch (t) {
     case SVt_PVAV:
 	CALL (json_create_add_array (jc, (AV *) r));
@@ -983,7 +1055,25 @@ json_create_handle_ref (json_create_t * jc, SV * input)
 	CALL (json_create_add_object (jc, (HV *) r));
 	break;
 
+    case SVt_NV:
+    case SVt_PVNV:
+	STRICT_NO_SCALAR;
+	CALL (json_create_add_float (jc, r));
+	break;
+
+    case SVt_IV:
+    case SVt_PVIV:
+	STRICT_NO_SCALAR;
+	CALL (json_create_add_integer (jc, r));
+	break;
+
+    case SVt_PV:
+	STRICT_NO_SCALAR;
+	CALL (json_create_add_string (jc, r));
+	break;
+
     case SVt_PVMG:
+	STRICT_NO_SCALAR;
 	/* There are some edge cases with blessed references
 	   containing numbers which we need to handle correctly. */
 	if (SvIOK (r)) {
@@ -997,26 +1087,29 @@ json_create_handle_ref (json_create_t * jc, SV * input)
 	}
 	break;
 
-    case SVt_PVGV:
-	/* Completely untested. */
-	CALL (json_create_add_string (jc, r));
-	break;
-
-    case SVt_PV:
-	CALL (json_create_add_string (jc, r));
-	break;
-
     default:
 	CALL (json_create_handle_unknown_type (jc, r));
     }
     return json_create_ok;
 }
 
+/* In strict mode, if no object handlers exist, then we reject the
+   object. */
+
+#define REJECT_OBJECT(objtype)						\
+    json_create_user_message (jc, json_create_unknown_type,		\
+			      "Object cannot be "			\
+			      "serialized to JSON: %s",			\
+			      objtype);					\
+    return json_create_unknown_type;
+
+
 static INLINE json_create_status_t
 json_create_handle_object (json_create_t * jc, SV * input)
 {
     const char * objtype;
     SV * r;
+
     r = SvRV (input);
     /* The second argument to sv_reftype is true if we
        look it up in the object table, false
@@ -1095,6 +1188,9 @@ json_create_handle_object (json_create_t * jc, SV * input)
 		}
 #endif /* 0 */
 	    nothandled:
+		if (jc->strict) {
+		    REJECT_OBJECT(objtype);
+		}
 		CALL (json_create_handle_ref (jc, input));
 	    }
 	}
@@ -1127,8 +1223,16 @@ json_create_recursively (json_create_t * jc, SV * input)
     }
     if (SvROK (input)) {
 	/* We have a reference, so decide what to do with it. */
-	if (sv_isobject (input) && (jc->handlers || jc->obj_handler)) {
-	    CALL (json_create_handle_object (jc, input));
+	if (sv_isobject (input)) {
+	    if (jc->handlers || jc->obj_handler) {
+		CALL (json_create_handle_object (jc, input));
+	    }
+	    else if (jc->strict) {
+		REJECT_OBJECT (sv_reftype (SvRV (input), 1));
+	    }
+	    else {
+		CALL (json_create_handle_ref (jc, input));
+	    }
 	}
 	else {
 	    CALL (json_create_handle_ref (jc, input));
@@ -1252,26 +1356,22 @@ json_create_run (json_create_t * jc, SV * input)
     return jc->output;
 }
 
-/* Master routine, callers should only ever use this. Everything above
-   is only for the sake of "json_create" to use. */
+/* Entry point for "create_json_strict". */
+
+static INLINE SV *
+json_create_strict (SV * input)
+{
+    json_create_t jc = {0};
+    jc.strict = 1;
+    return json_create_run (& jc, input);
+}
+
+/* Entry point for "create_json". */
 
 static INLINE SV *
 json_create (SV * input)
 {
-    /* With all the options, this really needs to be blanked out. Thus
-       "buffer" is moved from being inside "jc" to being inside
-       "json_create_run" above. */
     json_create_t jc = {0};
-    /* Set up the default options. */
-
-    /* Floating point number format. */
-    jc.fformat = 0;
-    /* Escape slash. */
-    jc.escape_slash = 0;
-
-    jc.unicode_escape_all = 0;
-    jc.handlers = 0;
-    jc.type_handler = 0;
     return json_create_run (& jc, input);
 }
 
