@@ -1,15 +1,34 @@
 package JSON::Create;
 require Exporter;
 @ISA = qw(Exporter);
-@EXPORT_OK = qw/create_json/;
+@EXPORT_OK = qw/create_json create_json_strict/;
 %EXPORT_TAGS = (
     all => \@EXPORT_OK,
 );
 use warnings;
 use strict;
-our $VERSION = '0.09_02';
-require XSLoader;
-XSLoader::load ('JSON::Create', $VERSION);
+our $VERSION = '0.19_01';
+# Are we running as XS?
+our $noxs;
+$noxs = $ENV{JSONCreatePP};
+# Did the XS load OK?
+our $xsok;
+if (! $noxs) {
+    eval {
+	#    die;
+	require XSLoader;
+	XSLoader::load ('JSON::Create', $VERSION);
+	$xsok = 1;
+    };
+    if ($@) {
+	$xsok = 0;
+    }
+}
+if (! $xsok || $noxs) {
+#    print "Pure perl OK.\n";
+    require JSON::Create::PP;
+    JSON::Create::PP->import (':all');
+}
 
 sub set_fformat
 {
@@ -26,6 +45,7 @@ sub set_fformat
 	    $obj->set_fformat_unsafe (0);
 	}
 	else {
+	    print "$fformat\n";
 	    $obj->set_fformat_unsafe ($fformat);
 	}
     }
@@ -55,6 +75,15 @@ sub obj
     }
 }
 
+sub del
+{
+    my ($obj, @list) = @_;
+    my $handlers = $obj->get_handlers ();
+    for my $item (@list) {
+	delete $handlers->{$item};
+    }
+}
+
 sub validate
 {
     my ($obj, $value) = @_;
@@ -65,5 +94,16 @@ sub validate
     $obj->set_validate ($value);
 }
 
+sub new
+{
+    my ($class) = @_;
+    if ($xsok) {
+	return bless jcnew (), $class;
+    }
+    else {
+	return JSON::Create::PP->new ();
+    }
+}
+
 1;
 
